import { Injectable } from '@angular/core';
import { HttpClientModule, HttpClient, HttpHeaders, HttpErrorResponse } from '@angular/common/http';
import { repeat } from 'rxjs/operators';
import { timer } from 'rxjs/observable/timer';

import {Strudelapp} from './strudelapp';
import { TesService } from './tes.service';
import { Identity } from './identity';
import { BatchInterface} from './batchinterface';
import { LaunchDialogComponent } from './launch-dialog/launch-dialog.component';
import { MatDialog, MatDialogRef, MAT_DIALOG_DATA, MatDialogModule } from '@angular/material';
import { DomSanitizer} from '@angular/platform-browser';
import{ BehaviorSubject } from 'rxjs';


// This service is responsible for starting applications
// Along the way it opens new windows/tabs to
/** This service is repsonible for starting applications
Along the way it opens new windows/tabs to determine
    1) Any configuration parameters for the application itself (if the app suports This
    2) The batch command to submit the application to a cluster (if any)
The windows are "watched" for URL changes and parameters are extracted from the
query string
*/


@Injectable({
  providedIn: 'root'
})
export class SubmitAppService {
  private configwindow: any;
  private appparams: string;
  // private submitcmd: string;


  private windowWatcher: any;
  private watchingForSubmit: boolean;
  public configApp: BehaviorSubject<any>;
  public appData: BehaviorSubject<any>;
  public submitcmd: BehaviorSubject<string>;
  public identity: BehaviorSubject<Identity>;
  public app: BehaviorSubject<Strudelapp>;
  public readyToLaunch: BehaviorSubject<Boolean>;


  constructor(private tesService: TesService,
              public dialog: MatDialog,
              private domSanitizer: DomSanitizer,
              ) {
    // this.appparams = null;
    // // this.submitcmd = null;
    // this.watchingForSubmit = false;
    this.app = new BehaviorSubject(null);
    this.appData = new BehaviorSubject(null);
    this.submitcmd = new BehaviorSubject(null);
    this.identity = new BehaviorSubject(null);
    this.readyToLaunch = new BehaviorSubject(false);
    this.app.subscribe(() => this.clearInfo());
    this.identity.subscribe(() => this.clearInfo());
    this.submitcmd.subscribe(() => this.canILaunch());
    this.appData.subscribe(()=> this.canILaunch());
  }

  launch() {
    let bi = new BatchInterface();
    bi.submitcmd = this.submitcmd.value;
    if (this.appData.value != null){ 
        //      let appparams = JSON.stringify(this.appData.value);
      let appparams = this.appData.value;
      this.tesService.submit(this.app.value,this.identity.value,bi,appparams);
    } else {
        this.tesService.submit(this.app.value,this.identity.value,bi);
    }
  }

  clearInfo() {
      this.appData.next(null);
      this.submitcmd.next(null);
  }

  canILaunch() {
    console.log('in submitAppService, evaluatin can I launch');
    if (this.submitcmd.value != null) {
        if (this.app.value.url == null || this.appData.value != null ) {
            this.readyToLaunch.next(true);
            return;
        } 
    }
    this.readyToLaunch.next(false);
  }

  // public launchAppModal(app: Strudelapp, identity: Identity) {
  //   let dialogRef = this.dialog.open(LaunchDialogComponent, {
  //     width: '600px',
  //     height: '300px',
  //     data: { batchcmdsafeurl: this.domSanitizer.bypassSecurityTrustResourceUrl(identity.site.url),
  //             app: app.name}
  //     })
  // }

  // public launchApp(app: Strudelapp, identity: Identity) {
  //   if (!(app.url === null) && !(app.url =='')) {
  //     this.openWindow(app,app.url,identity);
  //   } else {
  //     this.appparams = JSON.stringify([]);
  //     let redirect = window.location.toString();
  //     redirect = redirect.replace(/launch/,'finishlaunch');
  //     let paramstr = this.buildParams(app,identity,redirect);
  //     let windowloc = identity.site.url+'configure/'+app.name+"?"+paramstr;
  //     this.openWindow(app,windowloc,identity);
  //   }
  //   // this.openLaunchWindow(app, identitiy);
  // }
  //
  //
  // public openWindow(app: any,url: string, identity: Identity) {
  //   console.log('in openLaunchDialog');
  //   console.log(app);
  //
  //   let headers = new HttpHeaders();
  //   let options = { headers: headers, withCredentials: true};
  //   let redirect = window.location.toString();
  //   redirect = redirect.replace(/launch/,'finishlaunch');
  //   let paramstr = this.buildParams(app,identity,redirect);
  //   this.configwindow = window.open(url+"?"+paramstr);
  //   if (!(this.windowWatcher === null) && !(this.windowWatcher === undefined)) {
  //     this.windowWatcher.unsubscribe()
  //   }
  //   this.windowWatcher = timer(500).pipe(repeat()).subscribe(() => this.watchWindow(this.configwindow, identity, app,redirect));
  //   this.configwindow.focus();
  //
  // }
  //
  // public watchWindow(configwindow, id, app, redirect) {
  //   // Watch another window for url changes
  //   // if the url changes to include the submit parameters, submit the job
  //   // if the url changes to include appparameters, save them and open
  //   // a new window to get the submit parameters
  //   var location: string;
  //   try {
  //     location = configwindow.location.href;
  //     // if we can read the location and it contains a submitcmd arameter save it
  //     // and close the window
  //     if (configwindow.closed) { // if the window is closed without our help, presumable the user wanted to cancel the launch
  //       return
  //     }
  //     let match = location.match(/submitcmd\=([^&]+)(&|$)/);
  //     if (match) {
  //         this.submitcmd = decodeURIComponent(match[1]);
  //         if (!(configwindow.closed)) {
  //           configwindow.close()
  //         }
  //         this.windowWatcher.unsubscribe();
  //     }
  //     // if we can read the location and it contains an appparams parameter, save it
  //     // and close the window
  //     match = location.match(/appparams\=([^&]+)(&|$)/);
  //     if (match) {
  //         this.appparams = decodeURIComponent(match[1]);
  //         if (!(configwindow.closed)) {
  //           configwindow.close()
  //         }
  //         this.windowWatcher.unsubscribe();
  //     }
  //     // if we can read the location and it contains the redireURI, close the window
  //     let idx = location.indexOf(redirect);
  //     if (idx == 0) {
  //       configwindow.close();
  //     }
  //   } catch {
  //     location = null;
  //   }
  //   if (configwindow.closed) {
  //     this.windowWatcher.unsubscribe();
  //     // if the window is closed, see if we need to override appparams from localStorage
  //     try {
  //       let appparams = localStorage.getItem('appparams');
  //       if (appparams != null) {
  //         localStorage.removeItem('appparams');
  //         this.appparams = appparams;
  //       }
  //     } catch {
  //     }
  //     // if the window is closed and we have an appparams value but not a submitcmd
  //     // open a new window to get the submitcmd
  //
  //
  //     if (!(this.appparams === null) && (this.submitcmd === null)) {
  //       let paramstr = this.buildParams(app,id,redirect);
  //       let windowloc = id.site.url+'configure/'+app.name+"?"+paramstr;
  //       this.windowWatcher.unsubscribe();
  //       this.openWindow(app,windowloc,id);
  //     }
  //     // if the window is closed and we have an appparams valud AND a submitcmd
  //     // submit
  //     if (!(this.appparams == null) && !(this.submitcmd == null )) {
  //
  //       let bi = new BatchInterface();
  //       bi.submitcmd = this.submitcmd;
  //       this.tesService.submit(app,id,bi,this.appparams);
  //       this.appparams = null;
  //       this.submitcmd = null;
  //     }
  //   }
  // }
  //
  // private buildParams(app: Strudelapp, identity: Identity, redirect): string {
  //   let params = new URLSearchParams();
  //   let id = identity.copy_skip_catalog();
  //   id.site.appCatalog = null;
  //
  //   params.set('app',JSON.stringify(app));
  //   params.set('identity',JSON.stringify(id));
  //   params.set('redirect',redirect);
  //   return params.toString();
  // }

}
